#include <sdcc-lib.h>
#include <pic16f648a.h>	// /usr/share/sdcc/include/pic/pic16f648a.h
#include "inline.h"

//__code unsigned char at 0x2000 id_location[] = {'A', 'B', 'C', 'D', 0, 0, 0 };

unsigned int at 0x2007 __config =
	_CP_OFF &
	_DATA_CP_OFF &
	_LVP_OFF &
	_BOREN_OFF &	// for 3V
	_MCLRE_OFF &
	_PWRTE_OFF &
	_WDT_OFF &
	_INTRC_OSC_NOCLKOUT;

__code unsigned char at 0x2100 eedata[] = {'d', 'u', 'm', 'm', 'y', 0 };

#define LED_BIT		3

void initSerial();
unsigned char receiveSerial();
void transmitSerial(unsigned char c);

void forCalibOsc();

volatile unsigned char waitCntS;
volatile unsigned char waitCntM;
volatile unsigned char waitCntL;

void waitS();
void waitM();
void waitL();

#define waitSS(x) LD_A(x); waitS();
#define waitMM(x) LD_A(x); waitM();
#define waitLL(x) LD_A(x); waitL();

// TIMER_STEP: 5s=150, 1min=1800, 5min=9000
#define TIMER_STEP 9000
#define MAXSIG 16

void Intr(void) __interrupt 0 {  
}

void main(void) {

	unsigned char signal[MAXSIG];
	unsigned char sig_span;
	unsigned char sig_p;
	unsigned char sig_b;
	unsigned char off_timer;
	unsigned char last_tmr0;
	unsigned short timer_s;

/*			  _______
	NC		-|2		1|-		NC
	LED		-|3		0|-		NC
	NC		-|4		7|-		NC
	NC		-|5		6|-		NC
	GND		-|		 |-		Vcc
	IrSig	-|0		7|-		RelayA
	<Vcc>	-|1		6|-		RelayA
	TX		-|2		5|-		RelayB
	NC		-|3_____4|-		RelayB
*/

				//  76543210
	PORTA		= 0x00000000;		// RA5: Input only
	CMCON		= 0b00000111;
	TRISA		= 0b00100000;		// 1: 入力, 0: 出力

	PORTB		= 0b00000000;
	TRISB		= 0b00000011;		// 1: 入力, 0: 出力

	OPTION_REG	= 0b10010111;		// 初期値:11111111b
						// 7:PORTBプルアップ(0:使用する)
						// 6:RB0割り込みポート変化選択(0:立下り)	// IRセンサ: Active Low
						// 5:TMR0クロックソース選択ビット(0:内部)
						// 4:TMR0加算タイミング(0:立上り)
						// 3:プリスケーラ割り当てビット(0:TMR0)
						// 2~0プリスケーラレート選択ビット(011:1/16, 110:1/128, 111:1/256)
						// 4MHz => 1MHz = 1,000,000 / 256 / 256 = 15Hz

	INTCON		= 0b00010000;		// 初期値:0000000xb
						// 7:GIE  Global Interrupt Enable bit
						// 6:PEIE Peripheral Interrupt Enable bit
						// 5:T0IE TMR0 Overflow Interrupt Enable bit
						// 4:INTE RB0/INT External Interrupt Enable bit
						// 3:RBIE RB Port Change Interrupt Enable bit
						// 2:T0IF TMR0 Overflow Interrupt Flag bit
						// 1:INTF RB0/INT External Interrupt Flag bit
						// 0:RBIF RB Port Change Interrupt Flag bit

//	PIE1		= 0b00000000;		// 初期値:00000000b
//	PIR1		= 0b00000000;		// 初期値:00000000b
						// 7:EEIE EE Write Complete Interrupt Enable Bit
						// 6:CMIE Comparator Interrupt Enable bit
						// 5:RCIE USART Receive Interrupt Enable bit
						// 4:TXIE USART Transmit Interrupt Enable bit
						// 2:CCP1IE CCP1 Interrupt Enable bit
						// 1:TMR2IE TMR2 to PR2 Match Interrupt Enable bit
						// 0:TMR1IE TMR1 Overflow Interrupt Enable bit

//	forCalibOsc();

	initSerial();

	off_timer = 0;
	last_tmr0 = 0;
	timer_s = TIMER_STEP;

	/*------------------------------------------------------------------------------
	 *
	 *	メイン
	 *
	 */
	while(1) {

		for(sig_p = 0; sig_p < MAXSIG; sig_p++)	// clear signal buf
			signal[sig_p] = 0;

		sig_b = 0; sig_p = 0;

		if(off_timer) {							// falling edge を待ちつつ TMR0 を数える

			while(PORTB & 0b00000001) {			// wait Falling edge

				if((TMR0 ^ last_tmr0) & 0b10000000) {
					if(--timer_s == 0) {
						timer_s = TIMER_STEP;
						off_timer--;
					}
				}
				last_tmr0 = TMR0;

				if(!off_timer) {
					PORTB |= 0b11000000;		// relay Off
					waitMM(40);					// relay wait 20ms
					PORTB &= 0b00001111;
					waitLL(5);					// wait 0.5s

					INTF = 0;
	_asm
		SLEEP				;							SLEEP					; into Power-Down mode
		nop				;							NOP
	_endasm;
					break;
				}
			}

		} else {

			INTF = 0;
	_asm
		SLEEP				;							SLEEP					; into Power-Down mode
		nop				;							NOP
	_endasm;
		}

		while(1) {
			while(!(PORTB & 0b00000001)); 		// wait Raising edge

			sig_span = 0;
			while(PORTB & 0b00000001) {			// wait Falling edge
				waitSS(5);
				if(++sig_span > 200) break;
			}

			signal[sig_p] <<= 1;
			if(sig_span > 30)
				signal[sig_p] |= 1;				// check Active (long span) signal

			if(++sig_b == 8) {
				sig_b = 0;
				sig_p++;
			}

			if(sig_span > 200 || sig_p == MAXSIG) break;
		}

		for(sig_p = 0; sig_p < MAXSIG; sig_p++)	// for Serial debug
			transmitSerial(signal[sig_p]);

		if(signal[0] == 0xa0) {
			switch(signal[4]) {

			  case 0x04:						// IR signal '1' for ON
				off_timer = 0;					// off timer disable
				PORTB |= 0b00110000;			// relay On
				break;

			  case 0x44:						// IR signal '2' for Timer
				off_timer += 2;					// off timer += 10min
				for(sig_p = 0; sig_p < off_timer; sig_p++) {
					waitLL(3);
					PORTA |= 0b00001000;		// brink LED
					waitLL(2);
					PORTA &= 0b11110111;
				}
				break;

			  case 0x24:						// IR signal '3' for OFF
				off_timer = 0;					// off timer disable
				PORTB |= 0b11000000;			// relay Off
				break;

//			  case 0x64:						// IR signal '4' for ???
//			  case 0x14:						// IR signal '5' for ???
//			  case 0x54:						// IR signal '6' for ???
			}
		}

		waitMM(40);								// relay wait 20ms
		PORTB &= 0b00001111;

		waitLL(5);								// wait 0.5s
	}
}

/*------------------------------------------------------------------------------
 *
 *	シリアル初期設定
 *
 */
void initSerial() {
									// Baud Rate = FOSC / (16 * (X + 1))
									// 9615.38 = 4,000,000 / (16 * (25 + 1))

	SPBRG	= 25;					// Baud Rate Generator
	BRGH	= 1;					// High Baud Rate Select(High/Low)
	SYNC	= 0;					// USART Mode Select(Synchronous/Asynchronous)

	SPEN	= 1;					// Serial Port Enable
	CREN	= 1;					// Continuous Receive Enable

	TXEN	= 1;					// Transmit Enable
}

/*------------------------------------------------------------------------------
 *
 *	シリアル受信処理
 *		データ受信までブロックされる
 *		受信データをAレジスタに入れて返る
 *
 */
unsigned char receiveSerial() {

	unsigned char recv;

	while(1) {
		while(!RCIF);							// 受信フラグチェック

		if(OERR) {								// オーバランエラーチェック
			CREN = 0; CREN = 1;					// 受信回路リセット(OERRクリア)
		}
		if(FERR) {								// フレーミングエラーチェック
			recv = RCREG;						// ゴミを吸い取る
			continue;
		}
		break;
	}
	recv = RCREG;								// エラーなし、受信データを得る 
	return(recv);
}

unsigned char receiveSerial2() {				// ノンブロック

	unsigned char recv;

	if(!RCIF)									// 受信フラグチェック
		return(0x39);

	while(1) {
		if(OERR) {								// オーバランエラーチェック
			CREN = 0; CREN = 1;					// 受信回路リセット(OERRクリア)
		}
		if(FERR) {								// フレーミングエラーチェック
			recv = RCREG;						// ゴミを吸い取る
			return(0x38);
		}
		break;
	}
	recv = RCREG;								// エラーなし、受信データを得る 
	return(recv);
}
/*
		BIT	OERR, (RCSTA)		; オーバランエラーチェック
		JP	NZ, ovRunErr
		BIT	FERR, (RCSTA)		; フレーミングエラーチェック
		JP	NZ, frameErr
		LD	A, (RCREG)		; エラーなし、受信データを得る
		RET
ovRunErr
		RES	CREN, (RCSTA)
		SET	CREN, (RCSTA)		; 受信回路リセット(OERRクリア)
		LD	A, 0E1h			; '○'
		BIT	FERR, (RCSTA)		; フレーミングエラーチェック
		RET	Z
frameErr
		LD	A, (RCREG)		; ゴミを吸い取る
		LD	A, 87h			; '□'
		RET
*/

void transmitSerial(unsigned char c) {

		while(!TRMT);
		TXREG = c;								// USART Transmit data register
}

/*------------------------------------------------------------------------------
 *
 *	20秒周期LED点滅処理(クロック速度計測用)
 *
 */
void forCalibOsc() {
	_asm
forCalibOsc:					;			forCalibOsc:
		bsf	PORTA, LED_BIT		;							SET		LED_BIT, (PORTA)
						;	
		movlw	100			;							LD		A, 100					; 100を指定すると、
		call	WAITnx01sm25u		;							CALL	WAITnx01sm25u			; 2行で 100 x 100ms - 25us 消費する
		movlw	6			;							LD		A, 6					; 10を指定すると、
		call	WAITnx3p5u		;							CALL	WAITnx3p5u				; 2行で 6 x 3 + 5 = 23us 消費
						;	
		bcf	PORTA, LED_BIT		;							RES		LED_BIT, (PORTA)
						;	
		movlw	100			;							LD		A, 100					; 100を指定すると、
		call	WAITnx01sm25u		;							CALL	WAITnx01sm25u			; 2行で 100 x 100ms - 25us 消費する
		movlw	6			;							LD		A, 6					; 10を指定すると、
		call	WAITnx3p5u		;							CALL	WAITnx3p5u				; 2行で 6 x 3 + 5 = 23us 消費
						;	
		goto	forCalibOsc		;							JP		forCalibOsc
	_endasm;
}

void waitS() {
	_asm
						;			;-------------------------------------------------------------------------------
						;			;
						;			;	時間待ち(小)
						;			;
						;			;	    * 4MHz 時
						;			;				LD		A, 10					; 10を指定すると、
						;			;				CALL	WAITnx3p5u				; 2行で 10 x 3 + 5 = 35us 消費
						;			;
WAITnx3p5u:					;			WAITnx3p5u:
		movwf	_waitCntS		;							LD		(_waitCntS), A
WAITnx3p5u1:					;			WAITnx3p5u1:
		decfsz	_waitCntS		
		goto	WAITnx3p5u1		;							DJNZ	(_waitCntS), WAITnx3p5u1
						;			;				RET
	_endasm;
}

void waitM() {
	_asm
						;			;-------------------------------------------------------------------------------
						;			;
						;			;	時間待ち(中)
						;			;
						;			;	    * 4MHz 時
						;			;				LD		A, 20					; 0.01秒が欲しい場合、20を指定すると、
						;			;				CALL	WAITnx500m35u			; 2行で 20 x 500 - 35 = 9.965ms 消費する
						;			;
WAITnx500m35u:					;			WAITnx500m35u:
		movwf	_waitCntM		;							LD		(_waitCntM), A
		goto	WAITnx500m35u2		;							JP		WAITnx500m35u2
WAITnx500m35u1:	nop				;			WAITnx500m35u1:	NOP								;             =   1us
		movlw	12			;							LD		A, 12					;
		call	WAITnx3p5u		;							CALL	WAITnx3p5u				;  12 x 3 + 5 =  41us
WAITnx500m35u2:	movlw	150			;			WAITnx500m35u2:	LD		A, 150					;
		call	WAITnx3p5u		;							CALL	WAITnx3p5u				; 150 x 3 + 5 = 455us
		decfsz	_waitCntM		
		goto	WAITnx500m35u1		;							DJNZ	(_waitCntM), WAITnx500m35u1
						;			;				RET
	_endasm;
}

void waitL() {
	_asm
						;			;-------------------------------------------------------------------------------
						;			;
						;			;	時間待ち(大)
						;			;
						;			;	    * 4MHz 時
						;			;				LD		A, 10					; 1秒が欲しい場合、10を指定すると、
						;			;				CALL	WAITnx01sm25u			; 2行で 10 x 100ms - 25us 消費する
						;			;
WAITnx01sm25u:					;			WAITnx01sm25u:
		movwf	_waitCntL		;							LD		(_waitCntL), A
		goto	WAITnx01sm25u2		;							JP		WAITnx01sm25u2
WAITnx01sm25u1:	movlw	9			;			WAITnx01sm25u1:	LD		A, 9					;
		call	WAITnx3p5u		;							CALL	WAITnx3p5u				;   9 x   3 +  5 =    32us
WAITnx01sm25u2:	movlw	200			;			WAITnx01sm25u2:	LD		A, 200					;
		call	WAITnx500m35u		;							CALL	WAITnx500m35u			; 200 x 500 - 35 = 99965us
		decfsz	_waitCntL		
		goto	WAITnx01sm25u1		;							DJNZ	(_waitCntL), WAITnx01sm25u1
						;			;				RET
	_endasm;
}

